/*
** Copyright (C) 2007 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  ippair.c
**
**    A plug-in for rwfilter.
**
**    Reads a text file containing a list of IP pairs, then filter
**    records where the source and destination IP pairs match one of
**    the source and destination pairs in the text file.
**
**    Mark Thomas
**    September 2007
**
**    TODO:
**
**    Allow this to support CIDR ranges instead of just single IPs.
**    Would require reimplementing the tree structure, probably along
**    the lines of prefix maps.
*/

#include "silk.h"

RCSIDENT("$SiLK: ippair.c 9724 2007-12-05 20:04:54Z mthomas $");

#include "rwpack.h"
#include "dynlib.h"
#include "skvector.h"
#include "skstream.h"
#include "redblack.h"


/* EXPORTED FUNCTIONS */

/* The dynamic library glue code calls these */
int dynlib_api_version(void);
int setup(dynlibInfoStruct *dlISP, dynlibSymbolId appType);
void teardown(dynlibSymbolId appType);
void optionsUsage(dynlibSymbolId appType, FILE *fh);
int check(rwRec *rwrec);


/* DEFINES AND TYPEDEFS */

/*  The delimiter that must appear between the two IP blocks */
#define DELIM ' '

/*  The types of filtering we support and their count. */
#define IPPAIR_INDEX_COUNT 1

typedef enum {
    IPPAIR_ANY=0
    /* For now, don't worry about the order of the entries in the
     * file.  Eventually, we could make the file order dependent. */
} ippair_index_t;

/*
 *  Create an array of ip-pair-nodes and hand pointers the array's
 *  entries to redblack.  However, if we need more nodes, we can't use
 *  realloc() since that may make redblack's pointers invalid.
 *  Instead, we'll use a vector of arrays, where each array contains
 *  NODE_ARRAY_SIZE elements.
 */
#define NODE_ARRAY_SIZE 65536


/* an IP-pair node */
typedef struct _ippair {
    uint32_t    ip1;
    uint32_t    ip2;
} ippair_t;



/* LOCAL VARIABLES */

/* the name of this plugin */
static const char *pluginName = "ippair";

/* a vector to hold the ippair_t arrays */
static sk_vector_t *array_vec;

/* the current array we're working with and the number of elements it
 * contains */
static struct _cur_pair {
    ippair_t   *array;
    int         count;
} cur_pairs;


/* the redblack trees--one for each ippair index */
static struct rbtree *rb[IPPAIR_INDEX_COUNT];

/*
 * Options that get added to rwfilter.
 */
static struct option filterOptions[] = {
    {"ippair-any",      REQUIRED_ARG, 0, IPPAIR_ANY},
    {0, 0, 0, 0}        /* sentinel */
};

static char *filterOptionsHelp[] = {
    ("The (sIP,dIP) pair or (dIP,sIP) pair appears in the\n"
     "\tlist of IP-pairs read from the named text file"),
    (char*)NULL
};


/* LOCAL FUNCTION PROTOTYPES */

static int compare(const void *pa, const void *pb, const void *config);
static int growMemory(void);
static int initializeMemory(int ippair_idx);
static int optionsHandler(clientData cData, int opt_index, char *opt_arg);
static int parsefile(int ippair_idx, const char *path);


/* FUNCTION DEFINITIONS */


/*
 *  int dynlib_api_version(void);
 *
 *    Return the dynlib API version that this plugin was compiled
 *    against.
 */
int dynlib_api_version(void)
{
    return DYNLIB_API_VERSION;
}


/*
 *  setup
 *      Called by dynlib interface code to setup this plugin.  This
 *      routine should set up options handling, if required.
 *  Arguments:
 *      --pointer to the dynamic library interface structure
 *      --Application type that is using this plugin.
 *  Returns:
 *      DYNLIB_FAILED - if processing fails
 *      DYNLIB_WONTPROCESS - if application should do normal output
 *      DYNLIB_WILLPROCESS - if this plugin takes over output
 *  Side Effects:
 */
int setup(dynlibInfoStruct *dlISP, dynlibSymbolId appType)
{
    /* verify same number of options and help strings */
    assert((sizeof(filterOptions)/sizeof(struct option))
           == (sizeof(filterOptionsHelp)/sizeof(char*)));

    /* Make the plugin use the application's context, so we can add
     * options to main app. */
    skAppContextSet(dynlibGetAppContext(dlISP));

    /* initialize data */
    memset(rb, 0, sizeof(rb));
    array_vec = NULL;

    /* this plug-in only supports rwfilter */
    switch (appType) {
      case DYNLIB_SHAR_FILTER:
      case DYNLIB_EXCL_FILTER:
        if (optionsRegister(filterOptions, &optionsHandler, (clientData)dlISP))
        {
            skAppPrintErr("Unable to register options for %s plugin",
                          pluginName);
            return 1;
        }
        break;

      default:
        skAppPrintErr("Cannot use %s plug-in with %s application",
                      pluginName, skAppName());
        return 1;
    }

    /* we're only a pass/fail plug-in */
    return DYNLIB_WONTPROCESS;
}


/*
 *  teardown
 *      Called by dynlib interface code to tear down this plugin.
 *  Arguments:
 *      --Application type that is using this plugin
 *  Returns:
 *      None.
 */
void teardown(dynlibSymbolId UNUSED(appType))
{
    static int teardownFlag = 0;
    size_t i;
    void *a;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    /* destroy the redblack trees */
    for (i = 0; i < IPPAIR_INDEX_COUNT; ++i) {
        if (rb[i]) {
            rbdestroy(rb[i]);
            rb[i] = NULL;
        }
    }

    /* destroy the vector of array of nodes */
    if (array_vec) {
        for (i = 0; i < skVectorGetCount(array_vec); ++i) {
            skVectorGetValue(&a, array_vec, i);
            free(a);
        }
        skVectorDestroy(array_vec);
        array_vec = NULL;
    }

    return;
}


/*
 *  optionsUsage
 *      Called by dynlib interface code to allow this plugin to print
 *      the options it accepts.
 *  Arguments:
 *      --Application type that is using this plugin
 *  Returns:
 *      None.
 *  Side Effects:
 *      None.
 */
void optionsUsage(dynlibSymbolId UNUSED(appType), FILE *fh)
{
    int i;

    for (i = 0; filterOptions[i].name != NULL; ++i) {
        fprintf(fh, "--%s %s. %s\n", filterOptions[i].name,
                SK_OPTION_HAS_ARG(filterOptions[i]), filterOptionsHelp[i]);
    }
}


/*
 *  optionsHandler
 *      Called by options parser to handle one user option
 *  Arguments:
 *      clientData cData: the dynamic lib interface struct pointer
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int optionsHandler(clientData cData, int opt_index, char *opt_arg)
{
    /* Make sure option in range; else a fatal programmer error */
    if (opt_index < 0 || opt_index >= IPPAIR_INDEX_COUNT) {
        skAppPrintErr("Option %d not handled at %s:%d",
                      opt_index, __FILE__, __LINE__);
        exit(EXIT_FAILURE);
    }

    /* Verify that we have a file name */
    if (opt_arg == NULL || opt_arg[0] == '\0') {
        skAppPrintErr("Expected name of %s file",
                      filterOptions[opt_index].name);
        return 1;
    }

    if (initializeMemory(opt_index)) {
        return 1;
    }

    if (parsefile(opt_index, opt_arg)) {
        return 1;
    }

    /* Tell the dynlib glue-code that we are active. */
    dynlibMakeActive((dynlibInfoStruct*)cData);

    return 0;
}


/*
 *  check
 *      Called by dynlib interface code to filter records
 *  Arguments:
 *      a RW record
 *  Returns:
 *      0 to accept the record, 1 to reject the record
 */
int check(rwRec *rwrec)
{
    ippair_t node;
    uint32_t tmp_ip;

    /* look for sIP,dIP */
    node.ip1 = rwrec->sIP.ipnum;
    node.ip2 = rwrec->dIP.ipnum;

    if (rbfind(&node, rb[IPPAIR_ANY]) != NULL) {
        /* found it */
        return 0;
    }

    /* look for dIP,sIP */
    tmp_ip = node.ip1;
    node.ip1 = node.ip2;
    node.ip2 = tmp_ip;
    if (rbfind(&node, rb[IPPAIR_ANY]) != NULL) {
        /* found it */
        return 0;
    }

    /* no match */
    return 1;
}


/*
 *  gt_lt_eq = compare(cp_a, cp_b, config);
 *
 *    Comparison function for the redblack tree; compares the
 *    ippair_t values 'cp_a' and 'cp_b'.
 */
static int compare(const void *va, const void *vb, const void UNUSED(*config))
{
    ippair_t *a = (ippair_t*)va;
    ippair_t *b = (ippair_t*)vb;

    if (a->ip1 < b->ip1) return -1;
    if (a->ip1 > b->ip1) return  1;
    if (a->ip2 > b->ip2) return -1;
    if (a->ip2 < b->ip2) return  1;
    return 0;
}


/*
 *  ok = growMemory();
 *
 *    Create a new array of ippair_t's and put a pointer to that
 *    array into the global 'array_vec'.  Return 0 on success, or -1
 *    for an allocation error.
 */
static int growMemory(void)
{
    ippair_t *new_mem;

    assert(array_vec != NULL);
    new_mem = calloc(NODE_ARRAY_SIZE, sizeof(ippair_t));
    if (new_mem == NULL) {
        return -1;
    }

    if (skVectorAppendValue(array_vec, &new_mem)) {
        free(new_mem);
        return -1;
    }

    cur_pairs.array = new_mem;
    cur_pairs.count = 0;

    return 0;
}


/*
 *  ok = initializeMemory(ippair_idx);
 *
 *    Create a redblack tree to hold information about the ip-pairs
 *    that are index by 'ippair_idx', an ippair_index_t.
 *
 *    In addition, if we haven't yet, create the global 'array_vec'
 *    that is used to store pointers to arrays of ippair_t's, and
 *    allocate an array and add it to the vector.
 *
 *    Return 0 on success, or non-zero on allocation failure.
 */
static int initializeMemory(int ippair_idx)
{
    /* initialize the binary tree for this ippair_index */
    assert(ippair_idx >= 0 && ippair_idx < IPPAIR_INDEX_COUNT);

    /* Create the main memory of nodes */
    if (array_vec == NULL) {
        array_vec = skVectorNew(sizeof(ippair_t**));
        if (array_vec == NULL) {
            skAppPrintErr("Insufficient memory to create vector");
            return -1;
        }

        /* add an array to it */
        if (growMemory()) {
            skVectorDestroy(array_vec);
            array_vec = NULL;
            return -1;
        }
    }

    /* Create the redblack tree */
    if (rb[ippair_idx] == NULL) {
        rb[ippair_idx] = rbinit(&compare, NULL);
        if (rb[ippair_idx] == NULL) {
            skAppPrintErr("Insufficient memory to create redblack tree");
            return -1;
        }
    }

    return 0;
}


/*
 *  ok = parsefile(ippair_idx, path);
 *
 *    Parse the file at 'path' that holds the data for the
 *    'ippair_idx'ed redblack tree.  Return 0 on success, or non-zero
 *    otherwise.
 */
static int parsefile(int ippair_idx, const char *path)
{
    char line_buf[1024];
    int lc = 0;
    int err_count = 0;
    uint32_t ip1;
    uint32_t ip2;
    char *sp1, *sp2;
    skstream_t *stream;
    int read_rv = SKSTREAM_OK;
    int rv;

    rv = dynlibOpenDataInputStream(&stream, SK_CONTENT_TEXT, path);
    if (rv == -1) {
        skAppPrintErr("Problem with input file %s", path);
        return 1;
    }
    if (rv == 1) {
        /* Ignore the file */
        return 0;
    }

    if ((rv = skStreamSetCommentStart(stream, "#"))) {
        goto END;
    }

    /* read until end of file */
    while ((read_rv = skStreamGetLine(stream, line_buf, sizeof(line_buf), &lc))
           != SKSTREAM_ERR_EOF)
    {
        switch (read_rv) {
          case SKSTREAM_OK:
            /* good, we got our line */
            break;
          case SKSTREAM_ERR_LONG_LINE:
            /* bad: line was longer than sizeof(line_buf) */
            skAppPrintErr("Input line %s:%d too long. ignored",
                          path, lc);
            continue;
          default:
            /* unexpected error */
            skStreamPrintLastErr(stream, read_rv, &skAppPrintErr);
            goto END;
        }

        /* find first non-whitespace character */
        sp1 = line_buf;
        while (isspace((int)*sp1)) {
            ++sp1;
        }

        /* find the separator between the two IP(s), convert it to
         * NUL, and find and the start of the second IP. */
        sp2 = strchr(sp1, DELIM);
        if (sp2 == NULL) {
            skAppPrintErr("Missing separator in ip-pair at %s:%d",
                          path, lc);
            ++err_count;
            continue;
        }
        *sp2 = '\0';
        ++sp2;
        while (isspace((int)*sp2)) {
            ++sp2;
        }

        /* parse the IPs */
        rv = skStringParseIP(&ip1, sp1);
        if (rv) {
            skAppPrintErr("Invalid IP '%s' at %s:%d",
                          sp1, path, lc);
            ++err_count;
            continue;
        }
        rv = skStringParseIP(&ip2, sp2);
        if (rv) {
            skAppPrintErr("Invalid IP '%s' at %s:%d",
                          sp2, path, lc);
            ++err_count;
            continue;
        }

        /* if we make it here we have two IPs */

        if (cur_pairs.count == NODE_ARRAY_SIZE) {
            if (growMemory()) {
                skAppPrintErr(("Ran out of memory adding ip-pairs\n"
                               "\tfrom %s.  Edit the file and try again"),
                              path);
                ++err_count;
                goto END;
            }
        }
        cur_pairs.array[cur_pairs.count].ip1 = ip1;
        cur_pairs.array[cur_pairs.count].ip2 = ip2;

        (void)rbsearch(&cur_pairs.array[cur_pairs.count], rb[ippair_idx]);
        /* could be smarter here about only incrementing when
         * we've added a new node */
        ++cur_pairs.count;
    }

  END:
    skStreamDestroy(&stream);
    /* return error code if rv is non-zero */
    if (rv != 0) {
        return -1;
    }
    /* return error code if we did not get to the end of the stream */
    if (read_rv != SKSTREAM_ERR_EOF) {
        return -1;
    }
    /* return error code if we encountered errors */
    if (err_count) {
        return -1;
    }
    /* return error code if we do not have any entries */
    if (array_vec == NULL
        || 0 == skVectorGetCount(array_vec)
        || 0 == cur_pairs.count)
    {
        skAppPrintErr("No valid IP pairs read from input file '%s'",
                      path);
        return -1;
    }
    return 0;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
